// (C) 2001-2024 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.

`timescale 1ns / 100ps
module  debouncer #(
        parameter            FILTER_WIDTH = 4,   // Delay in clock cycle(s), e.g. 4, 5, 6...
        parameter            INIT_BIT     = 1'b1 // High (1) or low (0)
        )
        (
         input   wire        i_rst_n         , // Reset, low active
         input   wire        i_clk           , // System clock
         input   wire [7:0]  i_clk_div       , // System clock division factor
         input   wire        i_framein       , // Frame input
         output  wire        o_frameout        // Frame output after de-bouncing
         );

//--------------------------------------------------
//Signal defination
//--------------------------------------------------
wire                         rst_n      ; // 
wire                         clk        ; // 
wire     [7:0]               clk_div    /* synthesis keep */;
reg      [7:0]               clk_div_cnt/* synthesis keep */;
reg                          clk_div_pls/* synthesis keep */;
wire                         framein    /* synthesis keep */;
reg      [FILTER_WIDTH-1:0]  fr_shifter /* synthesis keep */; 
reg                          frameout   /* synthesis keep */; 

//--------------------------------------------------
//Main function body
//--------------------------------------------------
//Input assignments
assign rst_n   = i_rst_n;
assign clk     = i_clk;
assign clk_div = i_clk_div;
assign framein = i_framein;

//Output assignments
assign o_frameout = frameout;

////////////////////////////////////////////////////////
// Clock division control
always @(posedge clk or negedge rst_n)
begin
  if(!rst_n)
    clk_div_cnt <= 8'd0;
  else if(clk_div_cnt == clk_div)
    clk_div_cnt <= 8'd0;
  else
    clk_div_cnt <= clk_div_cnt + 1'b1;
end

always @(posedge clk or negedge rst_n)
begin
  if(!rst_n)
    clk_div_pls <= 1'b0;
  else if(clk_div_cnt == clk_div)
    clk_div_pls <= 1'b1;
  else
    clk_div_pls <= 1'b0;
end

////////////////////////////////////////////////////////
// Left shifter control
always @(posedge clk or negedge rst_n)
begin
  if(!rst_n)
    fr_shifter <= {FILTER_WIDTH{INIT_BIT}};
  else if(clk_div_pls)
    fr_shifter <= {fr_shifter[FILTER_WIDTH-2:0], framein};
  else
    fr_shifter <= fr_shifter;
end

////////////////////////////////////////////////////////
// Filter output control
always @(posedge clk or negedge rst_n)
begin
  if(!rst_n)
    frameout <= INIT_BIT;
  else if(&fr_shifter == 1'b1)
    frameout <= 1'b1;
  else if(|fr_shifter == 1'b0)
    frameout <= 1'b0;
  else
    frameout <= frameout;
end

endmodule
//------------------------------------------------------------------------------
// end of file
//------------------------------------------------------------------------------