/*****************************************************************************
 * INTEL CONFIDENTIAL
 * Copyright 2024 Intel Corporation.  All Rights Reserved.
 *
 * The source code contained or described herein and all documents related
 * to the source code ("Material") are owned by Intel Corporation or its
 * suppliers or licensors. Title to the Material remains with Intel
 * Corporation or its suppliers and licensors. The Material contains trade
 * secrets and proprietary and confidential information of Intel or its
 * suppliers and licensors. The Material is protected by worldwide copyright
 * and trade secret laws and treaty provisions. No part of the Material may
 * be used, copied, reproduced, modified, published, uploaded, posted,
 * transmitted, distributed, or disclosed in any way without Intel's prior
 * express written permission.
 *
 * No license under any patent, copyright, trade secret or other intellectual
 * property right is granted to or conferred upon you by disclosure or
 * delivery of the Materials, either expressly, by implication, inducement,
 * estoppel or otherwise. Any license under such intellectual property rights
 * must be express and approved by Intel in writing.
 *****************************************************************************/
#include <stdio.h>
#include "io.h"
#include "alt_types.h"
#include "config.h"
#include "itl_i2c_interface.h"
#include "itl_fpc202.h"


static alt_u8 fpc202_reset[2] = {1, 1};     // Default High
static alt_u8 fpc202_lpmode[2] = {0, 0};    // Default Low

alt_u32 itl_fpc202_read_reg(alt_u16 reg, alt_u32 *data)
{
    return itl_i2c_read_reg(ITL_I2C_2_BUS, ITL_FPC202_ADDR_SELF, reg, data, 8);
}

alt_u32 itl_fpc202_write_reg(alt_u16 reg, alt_u32 data)
{
    return itl_i2c_write_reg(ITL_I2C_2_BUS, ITL_FPC202_ADDR_SELF, reg, data, 8);
}

alt_u32 itl_fpc202_out_read_reg(alt_u8 slave_addr, alt_u16 reg, alt_u32 *data)
{
    return itl_i2c_read_reg(ITL_I2C_2_BUS, slave_addr, reg, data, 8);
}

alt_u32 itl_fpc202_out_write_reg(alt_u8 slave_addr, alt_u16 reg, alt_u32 data)
{
    return itl_i2c_write_reg(ITL_I2C_2_BUS, slave_addr, reg, data, 8);
}

alt_u32 fpc202_set_output(fpc202_port_enum_t port,alt_u32 level)
{
    alt_u32 data = 0;
    alt_u32 mask = 0;
    alt_u32 ret  = 0;

    switch(port)
    {
        case P0_S0_A:
            mask |= 0x01;
            break;
        case P0_S0_B:
            mask |= 0x10;
            break;
        case P1_S0_A:
            mask |= 0x04;
            break;
        case P1_S0_B:
            mask |= 0x40;
            break;
        default:
            return ITL_FAIL;
    }

    ret = itl_fpc202_read_reg(ITL_FPC202_REG_OUT_AB_VALUE, &data);
    if(ret != ITL_SUCCESS)
    {
        return ret;
    }
    
    if(level)
    {
        data |= mask;
    }
    else
    {
        data &= (~mask);
    }

    printf("set port%d = %lu\n", port, level);
    return itl_fpc202_write_reg(ITL_FPC202_REG_OUT_AB_VALUE, data);
}

void sync_fpc202_to_reg()
{
    alt_u32 data = 0;
    alt_u32 reg_data = 0;
    alt_u32 mask = 0;

    reg_data = IORD_32DIRECT(ITL_SYS_INFO_BASE, ITL_SYSINFO_REG_QSFP_CONF_IN);

    //interrupt
    if(itl_fpc202_read_reg(ITL_FPC202_REG_IN_A, &data) == ITL_SUCCESS)
    {
        mask = ITL_QSFP_INTL_MASK;
        if(data & ITL_FPC202_QSFP0_INTL_MASK)
            reg_data |= mask;
        else
            reg_data &= (~mask);

        mask = ITL_QSFP_INTL_MASK<<ITL_QSFP1_SHIFT;
        if(data & ITL_FPC202_QSFP1_INTL_MASK)
            reg_data |= mask;
        else
            reg_data &= (~mask);    
    }

    //present
    if(itl_fpc202_read_reg(ITL_FPC202_REG_IN_B, &data) == ITL_SUCCESS)
    {
        mask = ITL_QSFP_MOD_PRSL_MASK;
        if(data & ITL_FPC202_QSFP0_PRESENT_MASK)
            reg_data |= mask;
        else
            reg_data &= (~mask);

        mask = ITL_QSFP_MOD_PRSL_MASK<<ITL_QSFP1_SHIFT;
        if(data & ITL_FPC202_QSFP1_PRESENT_MASK)
            reg_data |= mask;
        else
            reg_data &= (~mask);    
    }

    IOWR_32DIRECT(ITL_SYS_INFO_BASE, ITL_SYSINFO_REG_QSFP_CONF_IN, reg_data);
    return;
}

void sync_reg_to_fpc202()
{
    alt_u32 reg_data = 0;
    alt_u32 mask = 0;
    fpc202_port_enum_t port;
    
    reg_data = IORD_32DIRECT(ITL_SYS_INFO_BASE, ITL_SYSINFO_REG_QSFP_CONF_OUT);

    for(int i=0; i<2; i++)
    {
        //reset  P0/1_S0_OUTA/B        
        port = ((i==0)? P0_S0_A: P1_S0_A);
        mask = (ITL_QSFP_RESET_MASK << (i*ITL_QSFP1_SHIFT));
        if((reg_data & mask) && (fpc202_reset[i] == 0))
        {
            fpc202_reset[i] = 1;
            fpc202_set_output(port, 1);
        }
        else if(((reg_data & mask)==0) && (fpc202_reset[i] == 1))
        {
            fpc202_reset[i] = 0;
            fpc202_set_output(port, 0);
        }

        //LPMODE, P0/1_S0_OUTB
        port = ((i==0)? P0_S0_B: P1_S0_B);
        mask = (ITL_QSFP_LPMODE_MASK << (i*ITL_QSFP1_SHIFT));
        if((reg_data & mask) && (fpc202_lpmode[i] == 0))
        {
            fpc202_lpmode[i] = 1;
            fpc202_set_output(port, 1);
        }
        else if(((reg_data & mask)==0) && (fpc202_lpmode[i] == 1))
        {
            fpc202_lpmode[i] = 0;
            fpc202_set_output(port, 0);
        }
        
    }
}

void itl_fpc202_sync()
{    
    sync_fpc202_to_reg();
    sync_reg_to_fpc202();
}


